package ru.dbotthepony.mc.otm.block.entity;

import net.minecraft.core.BlockPos;
import net.minecraft.core.Direction;
import net.minecraft.nbt.CompoundTag;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.world.entity.player.Inventory;
import net.minecraft.world.entity.player.Player;
import net.minecraft.world.inventory.AbstractContainerMenu;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.level.Level;
import net.minecraft.world.level.block.state.BlockState;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.items.CapabilityItemHandler;
import net.minecraftforge.items.IItemHandler;
import ru.dbotthepony.mc.otm.matter.IMatterGridCell;
import ru.dbotthepony.mc.otm.matter.MatterGrid;
import ru.dbotthepony.mc.otm.matter.MatterRegistry;
import ru.dbotthepony.mc.otm.Registry;
import ru.dbotthepony.mc.otm.capability.IMatterHandler;
import ru.dbotthepony.mc.otm.capability.IPatternStorage;
import ru.dbotthepony.mc.otm.capability.MatteryMachineEnergyStorage;
import ru.dbotthepony.mc.otm.container.MatteryContainer;
import ru.dbotthepony.mc.otm.menu.MatterScannerMenu;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.math.BigDecimal;
import java.util.Collection;

public class BlockEntityMatterScanner extends BlockEntityMatteryPoweredWorker implements IMatterGridCell {
	private static final TranslatableComponent NAME = new TranslatableComponent("block.overdrive_that_matters.matter_scanner");

	public final MatteryContainer input_slot = new MatteryContainer(this::setChanged, 1);

	private final LazyOptional<IItemHandler> handler_resolver = LazyOptional.of(() -> input_slot.handler(
		(slot, stack) -> MatterRegistry.hasMatterValue(stack),
		(slot, amount, stack) -> is_idling
	));

	public BlockEntityMatterScanner(BlockPos p_155229_, BlockState p_155230_) {
		super(Registry.BlockEntities.MATTER_SCANNER, p_155229_, p_155230_);
		energy = new MatteryMachineEnergyStorage(this, MatteryMachineEnergyStorage.MachineType.WORKER, new BigDecimal(40000), new BigDecimal(400), new BigDecimal(400));
	}

	private boolean valid = true;

	@Nonnull
	@Override
	public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
		if (valid && cap == CapabilityItemHandler.ITEM_HANDLER_CAPABILITY)
			return handler_resolver.cast();

		return super.getCapability(cap, side);
	}

	@Override
	public void invalidateCaps() {
		super.invalidateCaps();
		valid = false;
	}

	@Override
	public void reviveCaps() {
		super.reviveCaps();
		valid = true;
	}

	@Override
	public void notifyPatternsChanged() {
		is_idling = false;
	}

	@Override
	public void onNeighbourMatterCell(BlockPos pos, Level level, Direction direction, IMatterGridCell cell) {
		is_idling = false;
	}

	@Override
	protected Component getDefaultDisplayName() {
		return NAME;
	}

	@Nullable
	@Override
	public AbstractContainerMenu createMenu(int containerID, Inventory inventory, Player ply) {
		return new MatterScannerMenu(containerID, inventory, this);
	}

	@Override
	public CompoundTag save(CompoundTag nbt) {
		nbt.put("work_slots", input_slot.serializeNBT());
		return super.save(nbt);
	}

	@Override
	public void load(CompoundTag nbt) {
		input_slot.deserializeNBT(nbt.get("work_slots"));
		super.load(nbt);
	}

	private static final BigDecimal BASE_CONSUMPTION = new BigDecimal("40");

	@Override
	@Nonnull
	protected BigDecimal getBaseConsumption() {
		return BASE_CONSUMPTION;
	}

	@Nonnull
	@Override
	protected MachineJobFinish onJobFinish(MachineJob job) {
		if (grid == null)
			return new MachineJobFinish(false, 100);

		ItemStack stack = job.stack();

		if (stack.isEmpty() || !MatterRegistry.hasMatterValue(stack))
			return new MachineJobFinish();


		Collection<IPatternStorage.PatternState> get_state = grid.findPattern(stack.getItem());
		IPatternStorage.PatternState find_state = null;

		for (IPatternStorage.PatternState state : get_state) {
			if (state.item() == stack.getItem() && state.research_percent() < 1d) {
				find_state = state;
			}
			// как бы не так
			// может быть это было сделано специально?
			//else if (state.item() == stack.getItem() && state.research_percent() >= 1d) {
			//	return null;
			//}
		}

		IPatternStorage.PatternState new_state;

		if (find_state != null) {
			new_state = new IPatternStorage.PatternState(stack.getItem(), find_state.research_percent() + 0.2d);
		} else {
			new_state = new IPatternStorage.PatternState(stack.getItem(), 0.2d);
		}

		if (grid.insertPattern(new_state, false, false)) {
			return new MachineJobFinish();
		}

		return new MachineJobFinish(false, 200);
	}

	@Nullable
	@Override
	protected MachineJob getNextJob() {
		if (grid == null)
			return null;

		ItemStack stack = input_slot.getItem(0);

		if (stack.isEmpty() || !MatterRegistry.hasMatterValue(stack))
			return null;

		Collection<IPatternStorage.PatternState> get_state = grid.findPattern(stack.getItem());
		IPatternStorage.PatternState find_state = null;

		for (IPatternStorage.PatternState state : get_state) {
			if (state.item() == stack.getItem() && state.research_percent() < 1d) {
				find_state = state;
			} else if (state.item() == stack.getItem() && state.research_percent() >= 1d) {
				return null;
			}
		}

		IPatternStorage.PatternState new_state;

		if (find_state != null) {
			new_state = new IPatternStorage.PatternState(stack.getItem(), find_state.research_percent() + 0.2d);
		} else {
			new_state = new IPatternStorage.PatternState(stack.getItem(), 0.2d);
		}

		if (grid.insertPattern(new_state, false, true)) {
			ItemStack copy = stack.copy();
			copy.setCount(1);
			stack.shrink(1);
			input_slot.setChanged();
			return new MachineJob(copy, MatterRegistry.getMatterValue(copy).doubleValue() * 80000d);
		}

		return null;
	}

	@Nullable
	@Override
	public IPatternStorage getPatternStorage() {
		return null;
	}

	private MatterGrid grid;

	@Override
	public void setLevel(Level p_155231_) {
		super.setLevel(p_155231_);

		if (grid == null)
			scheduleDiscoverNeighbours(getBlockPos(), p_155231_);
	}

	@Nullable
	@Override
	public MatterGrid getMatterGrid() {
		return grid;
	}

	@Nullable
	@Override
	public IMatterHandler getMatterHandler() {
		return null;
	}

	@Override
	public boolean isValidMatterCell() {
		return valid;
	}

	@Override
	public void setMatterGrid(MatterGrid grid) {
		this.grid = grid;
	}
}
