package ru.dbotthepony.mc.otm.container;

import net.minecraft.nbt.CompoundTag;
import net.minecraft.nbt.ListTag;
import net.minecraft.nbt.Tag;
import net.minecraft.world.SimpleContainer;
import net.minecraft.world.item.ItemStack;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.items.IItemHandler;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.function.Consumer;

public class MatteryContainer extends SimpleContainer {
	protected final Runnable watcher;
	protected boolean ignore_change_notifications = false;

	public MatteryContainer(Runnable watcher, int i) {
		super(i);
		this.watcher = watcher;
	}

	public MatteryContainer of(@Nullable CompoundTag tag) {
		if (tag == null)
			return this;

		MatteryContainer container = MatteryContainer.of(watcher, tag, getContainerSize());

		if (handler != null)
			container.handler = handler.cloneOf(container);

		return container;
	}

	public void deserializeNBT(@Nullable CompoundTag tag) {
		ignore_change_notifications = true;

		for (int i = 0; i < getContainerSize(); i++) {
			setItem(i, ItemStack.EMPTY);
		}

		if (tag == null) {
			ignore_change_notifications = false;
			return;
		}

		// нам не интересен размер

		if (tag.get("items") instanceof ListTag list)
			for (int i = 0; i < Math.min(list.size(), getContainerSize()); i++)
				if (list.get(i) instanceof CompoundTag get_tag)
					setItem(i, ItemStack.of(get_tag));

		ignore_change_notifications = false;
		setChanged();
	}

	public MatteryContainer of(@Nullable Tag tag) {
		if (tag == null)
			return this;

		if (tag instanceof CompoundTag compound)
			return of(watcher, compound, getContainerSize());

		return this;
	}

	public void deserializeNBT(@Nullable Tag tag) {
		if (tag == null)
			return;

		if (tag instanceof CompoundTag compound) {
			deserializeNBT(compound);
		}
	}

	public MatteryContainer(Runnable watcher, ItemStack... p_19152_) {
		super(p_19152_);
		this.watcher = watcher;
	}

	public ContainerIteratorItemStack stackIterator() {
		return new ContainerIteratorItemStack(this);
	}

	public void consumeNonEmpty(Consumer<ItemStack> consumer) {
		stackIterator().consume(consumer);
	}

	public <T> ContainerIteratorCapability<T> capabilityIterator(Capability<T> cap) {
		return new ContainerIteratorCapability<>(this, cap);
	}

	public <T> void consumeCapability(Capability<T> cap, Consumer<T> consumer) {
		capabilityIterator(cap).consume(consumer);
	}

	@Override
	public void setChanged() {
		if (ignore_change_notifications)
			return;

		super.setChanged();
		watcher.run();
	}

	public CompoundTag serializeNBT() {
		CompoundTag tag = new CompoundTag();
		tag.putInt("size", getContainerSize());
		ListTag list = new ListTag();
		tag.put("items", list);

		for (int i = 0; i < getContainerSize(); i++)
			list.add(getItem(i).serializeNBT());

		return tag;
	}

	public static MatteryContainer of(Runnable watcher, @Nullable CompoundTag tag, int fallback_size) {
		if (tag == null)
			return new MatteryContainer(watcher, fallback_size);

		MatteryContainer container;

		if (tag.contains("size"))
			container = new MatteryContainer(watcher, tag.getInt("size"));
		else
			container = new MatteryContainer(watcher, fallback_size);

		if (tag.get("items") instanceof ListTag list)
			for (int i = 0; i < Math.min(list.size(), container.getContainerSize()); i++)
				if (list.get(i) instanceof CompoundTag get_tag)
					container.setItem(i, ItemStack.of(get_tag));

		return container;
	}

	public boolean hasEmptySlot() {
		for (int i = 0; i < getContainerSize(); i++) {
			if (getItem(i).isEmpty())
				return true;
		}

		return false;
	}

	@FunctionalInterface
	public interface MatteryContainerInsertValidator {
		boolean apply(int slot, @Nonnull ItemStack stack);
	}

	@FunctionalInterface
	public interface MatteryContainerExtractValidator {
		boolean apply(int slot, int amount, @Nonnull ItemStack stack);
	}

	protected MatteryContainerHandler handler;

	public IItemHandler handler(MatteryContainerInsertValidator insert_validator, MatteryContainerExtractValidator extract_validator) {
		if (handler != null)
			return handler;

		return handler = new MatteryContainerHandler(this, insert_validator, extract_validator);
	}

	public IItemHandler handler(MatteryContainerInsertValidator insert_validator) {
		if (handler != null)
			return handler;

		return handler = new MatteryContainerHandler(this, insert_validator);
	}

	public IItemHandler handler() {
		if (handler != null)
			return handler;

		return handler = new MatteryContainerHandler(this);
	}

	public int getMaxStackSize(int slot) {
		return getMaxStackSize();
	}

	static class MatteryContainerHandler implements IItemHandler {
		private final MatteryContainer container;
		private final MatteryContainerInsertValidator insert_validator;
		private final MatteryContainerExtractValidator extract_validator;

		MatteryContainerHandler(MatteryContainer container, MatteryContainerInsertValidator insert_validator, MatteryContainerExtractValidator extract_validator) {
			this.container = container;
			this.insert_validator = insert_validator;
			this.extract_validator = extract_validator;
		}

		MatteryContainerHandler(MatteryContainer container) {
			this(container, (slot, stack) -> true, (slot, amount, stack) -> true);
		}

		MatteryContainerHandler(MatteryContainer container, MatteryContainerInsertValidator insert_validator) {
			this(container, insert_validator, (slot, amount, stack) -> true);
		}

		@Override
		public int getSlots() {
			return container.getContainerSize();
		}

		@Nonnull
		@Override
		public ItemStack getStackInSlot(int slot) {
			return container.getItem(slot);
		}

		@Nonnull
		@Override
		public ItemStack insertItem(int slot, @Nonnull ItemStack stack, boolean simulate) {
			if (!insert_validator.apply(slot, stack))
				return stack;

			ItemStack self_stack = container.getItem(slot);

			if (self_stack.isEmpty()) {
				if (!simulate) {
					container.setItem(slot, stack.copy());
				}

				return ItemStack.EMPTY;
			} else if (self_stack.isStackable() && self_stack.getMaxStackSize() > self_stack.getCount() && ItemStack.isSameItemSameTags(self_stack, stack)) {
				int new_count = Math.min(self_stack.getMaxStackSize(), self_stack.getCount() + stack.getCount());
				int diff = new_count - self_stack.getCount();

				if (diff != 0) {
					if (!simulate) {
						self_stack.grow(diff);
						container.setChanged();
					}

					ItemStack copy = stack.copy();
					copy.shrink(diff);
					return copy;
				}
			}

			return stack;
		}

		@Nonnull
		@Override
		public ItemStack extractItem(int slot, int amount, boolean simulate) {
			if (amount == 0)
				return ItemStack.EMPTY;

			if (amount < 0)
				throw new IllegalArgumentException("Can not extract negative amount of items");

			ItemStack self_stack = container.getItem(slot);

			if (self_stack.isEmpty())
				return ItemStack.EMPTY;

			if (!extract_validator.apply(slot, amount, self_stack))
				return ItemStack.EMPTY;

			int minimal = Math.min(amount, self_stack.getCount());
			ItemStack copy = self_stack.copy();
			copy.setCount(minimal);

			if (!simulate) {
				self_stack.shrink(minimal);
				container.setChanged();
			}

			return copy;
		}

		@Override
		public int getSlotLimit(int slot) {
			return container.getMaxStackSize();
		}

		@Override
		public boolean isItemValid(int slot, @Nonnull ItemStack stack) {
			return insert_validator.apply(slot, stack);
		}

		MatteryContainerHandler cloneOf(MatteryContainer container) {
			return new MatteryContainerHandler(container, insert_validator, extract_validator);
		}
	}
}
