package ru.dbotthepony.mc.otm.item;

import com.google.common.collect.ImmutableList;
import net.minecraft.ChatFormatting;
import net.minecraft.core.Direction;
import net.minecraft.nbt.*;
import net.minecraft.network.chat.Component;
import net.minecraft.network.chat.TranslatableComponent;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.world.item.CreativeModeTab;
import net.minecraft.world.item.Item;
import net.minecraft.world.item.ItemStack;
import net.minecraft.world.item.TooltipFlag;
import net.minecraft.world.level.Level;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.capabilities.ICapabilityProvider;
import net.minecraftforge.common.util.LazyOptional;
import net.minecraftforge.registries.ForgeRegistry;
import net.minecraftforge.registries.RegistryManager;
import ru.dbotthepony.mc.otm.capability.IPatternStorage;
import ru.dbotthepony.mc.otm.capability.MatteryCapability;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.util.*;

public class ItemPatternStorage extends Item {
	public final int capacity;

	public ItemPatternStorage(int capacity) {
		super(new Properties().tab(CreativeModeTab.TAB_MISC).stacksTo(1));
		this.capacity = capacity;
	}

	@Nullable
	@Override
	public ICapabilityProvider initCapabilities(ItemStack stack, @Nullable CompoundTag nbt) {
		return new ItemPatternStorageCapability(stack, capacity);
	}

	@Override
	public void appendHoverText(ItemStack p_41421_, @Nullable Level p_41422_, List<Component> list, TooltipFlag p_41424_) {
		p_41421_.getCapability(MatteryCapability.PATTERN).ifPresent(capability -> {
			list.add(new TranslatableComponent("otm.item.pattern.stored", capability.getStored(), capability.getCapacity()).withStyle(ChatFormatting.GRAY));

			for (IPatternStorage.PatternState state : capability.getStoredPatterns()) {
				list.add(new TranslatableComponent("otm.item.pattern.line", state.item().getName(new ItemStack(state.item(), 1)), String.format("%.2f", state.research_percent() * 100d)).withStyle(ChatFormatting.AQUA));
			}
		});

		super.appendHoverText(p_41421_, p_41422_, list, p_41424_);
	}

	public static class ItemPatternStorageCapability implements IPatternStorage, ICapabilityProvider {
		private final LazyOptional<IPatternStorage> resolver = LazyOptional.of(() -> this);
		public final ItemStack stack;
		public final int capacity;

		public ItemPatternStorageCapability(ItemStack stack, int capacity) {
			this.stack = stack;
			this.capacity = capacity;
		}

		@Override
		public int getCapacity() {
			return capacity;
		}

		@Override
		public int getStored() {
			return getStoredPatterns().size();
		}

		@Nonnull
		@Override
		public <T> LazyOptional<T> getCapability(@Nonnull Capability<T> cap, @Nullable Direction side) {
			if (cap == MatteryCapability.PATTERN)
				return resolver.cast();

			return LazyOptional.empty();
		}

		@Override
		public Collection<PatternState> getStoredPatterns() {
			CompoundTag tag = stack.getOrCreateTag();

			if (!tag.contains("otm_patterns") || !(tag.get("otm_patterns") instanceof ListTag))
				return Collections.EMPTY_LIST;

			ListTag list = tag.getList("otm_patterns", Tag.TAG_COMPOUND);
			ArrayList<PatternState> state_list = new ArrayList<>();

			for (int i = 0; i < list.size(); i++) {
				CompoundTag get_tag = list.getCompound(i);

				if (get_tag.get("id") instanceof StringTag id && get_tag.get("progress") instanceof DoubleTag progress) {
					Item get_item = RegistryManager.ACTIVE.getRegistry(Item.class).getValue(new ResourceLocation(id.getAsString()));

					if (get_item != null)
						state_list.add(new PatternState(get_item, progress.getAsDouble()));
				}
			}

			return ImmutableList.copyOf(state_list);
		}

		@Override
		public boolean insertPattern(PatternState pattern, boolean only_update, boolean simulate) {
			CompoundTag tag = stack.getOrCreateTag();

			ListTag list;

			if (tag.get("otm_patterns") instanceof ListTag list1) {
				list = list1;
			} else {
				if (only_update)
					return false;

				if (simulate)
					return capacity > 0;

				list = new ListTag();
				tag.put("otm_patterns", list);
			}

			String compare_with = Objects.requireNonNull(pattern.item().getRegistryName()).toString();

			int invalid_entries = 0;

			for (int i = 0; i < list.size(); i++) {
				CompoundTag get_tag = list.getCompound(i);

				if (get_tag.get("id") instanceof StringTag id) {
					if (compare_with.equals(id.getAsString())) {
						if (!simulate) {
							get_tag.putDouble("progress", pattern.research_percent());
						}

						return true;
					}
				} else {
					invalid_entries++;
				}
			}

			if (only_update || capacity <= list.size() - invalid_entries)
				return false;

			if (invalid_entries > 0) {
				if (simulate)
					return true;

				for (int i = 0; i < list.size(); i++) {
					CompoundTag get_tag = list.getCompound(i);

					if (!(get_tag.get("id") instanceof StringTag)) {
						get_tag.putString("id", compare_with);
						get_tag.putDouble("progress", pattern.research_percent());
						return true;
					}
				}
			}

			if (simulate)
				return true;

			CompoundTag get_tag = new CompoundTag();
			get_tag.putString("id", compare_with);
			get_tag.putDouble("progress", pattern.research_percent());
			list.add(get_tag);

			return true;
		}
	}
}
